#ifndef _KGR_PATCH_BSC1039496_LOCAL_H
#define _KGR_PATCH_BSC1039496_LOCAL_H

#include <asm/compat.h>

struct user_arg_ptr {
#ifdef CONFIG_COMPAT
	bool is_compat;
#endif
	union {
		const char __user *const __user *native;
#ifdef CONFIG_COMPAT
		const compat_uptr_t __user *compat;
#endif
	} ptr;
};

struct proc_maps_private {
	struct inode *inode;
	struct task_struct *task;
	struct mm_struct *mm;
#ifdef CONFIG_MMU
	struct vm_area_struct *tail_vma;
#endif
#ifdef CONFIG_NUMA
	struct mempolicy *task_mempolicy;
#endif
};

struct mm_slot {
	struct hlist_node hash;
	struct list_head mm_node;
	struct mm_struct *mm;
};

struct khugepaged_scan {
	struct list_head mm_head;
	struct mm_slot *mm_slot;
	unsigned long address;
};

struct proc_dir_entry {
	unsigned int low_ino;
	umode_t mode;
	nlink_t nlink;
	kuid_t uid;
	kgid_t gid;
	loff_t size;
	const struct inode_operations *proc_iops;
	const struct file_operations *proc_fops;
	struct proc_dir_entry *parent;
	struct rb_root subdir;
	struct rb_node subdir_node;
	void *data;
	atomic_t count;		/* use count */
	atomic_t in_use;	/* number of callers into module in progress; */
			/* negative -> it's going away RSN */
	struct completion *pde_unload_completion;
	struct list_head pde_openers;	/* who did ->open, but not ->release */
	spinlock_t pde_unload_lock; /* proc_fops checks and pde_users bumps */
	u8 namelen;
	char name[];
};

union proc_op {
	int (*proc_get_link)(struct dentry *, struct path *);
	int (*proc_show)(struct seq_file *m,
		struct pid_namespace *ns, struct pid *pid,
		struct task_struct *task);
};

struct proc_inode {
	struct pid *pid;
	int fd;
	union proc_op op;
	struct proc_dir_entry *pde;
	struct ctl_table_header *sysctl;
	struct ctl_table *sysctl_entry;
	const struct proc_ns_operations *ns_ops;
	struct inode vfs_inode;
};

#define kgr_pgd_offset_k(address) pgd_offset(kgr_init_mm, (address))
#define kgr_alloc_page_vma(gfp_mask, vma, addr)			\
	kgr_alloc_pages_vma(gfp_mask, 0, vma, addr, numa_node_id(), false)
#define kgr_alloc_hugepage_vma(gfp_mask, vma, addr, order)	\
	kgr_alloc_pages_vma(gfp_mask, order, vma, addr, numa_node_id(), true)
#define kgr__alloc_zeroed_user_highpage(movableflags, vma, vaddr) \
	kgr_alloc_page_vma(GFP_HIGHUSER | __GFP_ZERO | movableflags, vma, vaddr)

#define kgr_khugepaged_enabled()					       \
	(*kgr_transparent_hugepage_flags &				       \
	 ((1<<TRANSPARENT_HUGEPAGE_FLAG) |		       \
	  (1<<TRANSPARENT_HUGEPAGE_REQ_MADV_FLAG)))
#define kgr_khugepaged_always()				\
	(*kgr_transparent_hugepage_flags &			\
	 (1<<TRANSPARENT_HUGEPAGE_FLAG))
#define kgr_khugepaged_req_madv()					\
	(*kgr_transparent_hugepage_flags &				\
	 (1<<TRANSPARENT_HUGEPAGE_REQ_MADV_FLAG))
#define kgr_transparent_hugepage_use_zero_page()				\
	(*kgr_transparent_hugepage_flags &					\
	 (1<<TRANSPARENT_HUGEPAGE_USE_ZERO_PAGE_FLAG))
#define kgr_khugepaged_defrag()					\
	(*kgr_transparent_hugepage_flags &				\
	 (1<<TRANSPARENT_HUGEPAGE_DEFRAG_KHUGEPAGED_FLAG))

#define kgr_for_each_populated_zone(zone)		        \
	for (zone = (kgr_first_online_pgdat())->node_zones; \
	     zone;					\
	     zone = kgr_next_zone(zone))			\
		if (!populated_zone(zone))		\
			; /* do nothing */		\
		else

#if defined(CONFIG_DEBUG_VM_RB)
#error "kGraft patch supports only !CONFIG_DEBUG_VM_RB"
#endif
#define kgr_validate_mm(mm) do { } while (0)

#define kgr_free_mm(mm)	(kmem_cache_free(*kgr_mm_cachep, (mm)))

/* trace_sched_process_exec */
#define PARAMS(args...) args
#define TP_PROTO(args...)       args
#define TP_ARGS(args...)        args
#define TP_CONDITION(args...)   args

#undef __DECLARE_TRACE
#define __DECLARE_TRACE(name, proto, args, cond, data_proto, data_args) \
	extern struct tracepoint *kgr__tracepoint_##name;			\
	static inline void kgr_trace_##name(proto)				\
	{								\
		if (unlikely(static_key_count(&kgr__tracepoint_##name->key) > 0))               \
			__DO_TRACE(kgr__tracepoint_##name,		\
				TP_PROTO(data_proto),			\
				TP_ARGS(data_args),			\
				TP_CONDITION(cond),,);			\
		if (IS_ENABLED(CONFIG_LOCKDEP) && (cond)) {		\
			rcu_read_lock_sched_notrace();			\
			rcu_dereference_sched(kgr__tracepoint_##name->funcs);\
			rcu_read_unlock_sched_notrace();		\
		}							\
	}								\
	__DECLARE_TRACE_RCU(name, PARAMS(proto), PARAMS(args),		\
		PARAMS(cond), PARAMS(data_proto), PARAMS(data_args))	\
	static inline int						\
	kgr_register_trace_##name(void (*probe)(data_proto), void *data)	\
	{								\
		return tracepoint_probe_register(kgr__tracepoint_##name,	\
						(void *)probe, data);	\
	}								\
	static inline int						\
	kgr_register_trace_prio_##name(void (*probe)(data_proto), void *data,\
				   int prio)				\
	{								\
		return tracepoint_probe_register_prio(kgr__tracepoint_##name, \
					      (void *)probe, data, prio); \
	}								\
	static inline int						\
	kgr_unregister_trace_##name(void (*probe)(data_proto), void *data)	\
	{								\
		return tracepoint_probe_unregister(kgr__tracepoint_##name,\
						(void *)probe, data);	\
	}								\
	static inline void						\
	kgr_check_trace_callback_type_##name(void (*cb)(data_proto))	\
	{								\
	}								\
	static inline bool						\
	kgr_trace_##name##_enabled(void)					\
	{								\
		return static_key_count(&kgr__tracepoint_##name->key);	\
	}

#undef TRACE_EVENT
#define TRACE_EVENT(name, proto, args, struct, assign, print)   \
	DECLARE_TRACE(name, PARAMS(proto), PARAMS(args))

TRACE_EVENT(sched_process_exec,

	TP_PROTO(struct task_struct *p, pid_t old_pid,
		 struct linux_binprm *bprm),

	TP_ARGS(p, old_pid, bprm),

	TP_STRUCT__entry(
		__string(       filename,       bprm->filename  )
		__field(        pid_t,          pid             )
		__field(        pid_t,          old_pid         )
	),

	TP_fast_assign(
		__assign_str(filename, bprm->filename);
		__entry->pid            = p->pid;
		__entry->old_pid        = old_pid;
	),

	TP_printk("filename=%s pid=%d old_pid=%d", __get_str(filename),
		  __entry->pid, __entry->old_pid)
);

#endif
